% muc.pl - Verification of process specifications in mu-calculus.
% 
% (c) 2003 Martin Pitt <martin@piware.de>
%
% This software can be used freely under the terms and conditions of
% the GNU Lesser General Public License (LGPL). See
% http://www.gnu.org/copyleft/lgpl.html for details.
%
% -------------------------------------------------------------------
% Notational translation for formulas:
%
% CCL		    PROLOG
% ---		    ------
% Z		    z			    (atomic property or variable)
% not F		    ~F			    (negation)
% F1 /\ F2	    F1 /\ F2		    (conjunction)
% F1 v F2	    F1 \/ F2		    (disjunction)
%
% [A] F		    all([a1, a2,...], F)    (universal modal)
%  _
% [A] F		    allbut([a1, a2,...], F) (universal modal, [Delta - A])
% [] F		    all( F )		    (universal modal, [Delta])
%
% <A> F		    ex([a1, a2,...], F)	    (existential modal)
%  _
% <A> F		    exbut([a1, a2,...], F)  (existential modal, <Delta - A>)
% <> F		    ex( F )		    (existential modal, <Delta>)
%
% vZ. F		    nu( z, F )		    (greatest fixpoint)
% Z. F		    mu( z, F )		    (least fixpoint)
%
% -------------------------------------------------------------------
% These clauses must be defined (separately include them):
%   next( Process, Transition, Process' )
%   expand_proc_const( Constant, Definition ) 
%   has_property( Process, Property )
% ---
% Exported clauses:
%   failure( Formula, Process, Run ): 
%	succeeds if Formula does not hold for Process, with Run being
%	a counterexample
% -------------------------------------------------------------------

:- op( 100, fy, ~ ).   % formulas: negation
:- op( 550, yfx, \/ ).   % redefine to give \/ lower prio than /\


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% mu calculus tableau proof
%
% failure( Formula, Process, Run, Subst )
%
% search for a failing run Run of Formula in Process.
%
% Subst = [u1/Def_u1, ...]: Map (list of atom/definition) that
%			    contains the temporary substitutions
%			    for the fixpoint rules
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic( occurred_u/2 ).

clean :-
    abolish( occurred_u/2 ),
    asserta( (occurred_u(a,true) :- fail) ). % dummy clause since one must be present

failure( F, P, R ) :-
    failure( F, P, R, [] ).

% logical constants
failure( true, _, [], _ ) :- !, fail.
failure( false, P, [], _ ) :- !, printfail( false, P, [] ).

% expand process name constants
failure( F, K, R, S ) :-
    expand_proc_const( K, P ) -> !, failure( F, P, R, S );
    atom(K), write( 'Error: no process definition for ' ), write(K), nl. 

% U-rule
failure( U, P, R, S ) :- 
    atom( U ), map( U, S, Def ), !, Def =.. [Op,Z,Body],
    (
	occurred( U, P ) ->
	    ( Op == mu -> R=[], printfail( U, P, S ); fail );
	    subst( Body, Z, U, Body1 ), asserta(occurred_u(U, P)), failure( Body1, P, R, S )
    ).

% atomic property
failure( Z, P, [], S ) :-
    ( atom( Z ); Z = _^Z1, atom(Z1) ), (
	has_property( P, Z ) -> fail; 
	printfail( Z, P, S ) 
    ).

% negated atomic property
failure( ~Z, P, [], S ) :-
    ( atom( Z ); Z = _^Z1, atom(Z1) ), (
	has_property( P, Z ) -> printfail( ~Z, P, S );
	fail
    ).

% composite

failure( F /\ G, P, R, S ) :- failure( F, P, R, S ); failure( G, P, R, S ).

failure( F \/ G, P, R, S ) :- failure( F, P, _, S ), failure( G, P, R, S ).

failure( all( [A|As], F ), P, R, S ) :- 
    next( P, A, P1 ), failure( F, P1, R1, S ), R = [A|R1]; 
    failure( all( As, F ), P, R, S ).

failure( allbut( ExclA, F ), P, R, S ) :- 
    poss_trans_but( P, ExclA, T ),
    failure( all( T, F ), P, R, S ).

failure( all( F ), P, R, S ) :- 
    failure( allbut( [], F ), P, R, S ).

failure( ex( [], F ), P, _, S ) :- printfail( ex([],F), P, S ).

failure( ex( [A|As], F ), P, R, S ) :- 
    ( As = [] -> true; failure( ex( As, F), P, Rs, S ) ),
    ( next( P, A, P1 ) -> failure( F, P1, R1, S ); R1 = Rs ),
    R = [A|R1].

failure( exbut( ExclA, F ), P, R, S ) :- 
    poss_trans_but( P, ExclA, T ),
    failure( ex( T, F ), P, R, S ).

failure( ex( F ), P, R, S ) :- 
    failure( exbut( [], F ), P, R, S ).

failure( nu(Z, G), P, R, S ) :-
    create_atom( U ),
    failure( U, P, R, [U/nu(Z,G) | S] ).

failure( mu(Z, G), P, R, S ) :-
    create_atom( U ),
    failure( U, P, R, [U/mu(Z,G) | S] ).

failure( ~ F, P, R, S ) :-
    neg( F, FN ),
    failure( FN, P, R, S ).

failure( F, P, R, S ) :- 
    macro( F, Def ), 
    failure( Def, P, R, S ).

% negation

neg( ~F, F ).
neg( F /\ G, ~F \/ ~G ).
neg( F \/ G, ~F /\ ~G ).

neg( all(A,F), ex(A,~F) ).
neg( all(F), ex(~F) ).
neg( allbut(A,F), exbut(A,~F) ).

neg( ex(A,F), all(A,~F) ).
neg( ex(F), all(~F) ).
neg( exbut(A,F), allbut(A,~F) ).

neg( nu(Z,F), mu(Z,~F1) ) :- subst( F, Z, ~Z, F1 ).
neg( mu(Z,F), nu(Z,~F1) ) :- subst( F, Z, ~Z, F1 ).

neg( F, ~Def ) :- macro( F, Def ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Macros
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro( ag(F), nu(Z, F /\ all(Z) ) ):- create_atom(Z).
macro( af(F), mu(Z, F \/ ( ex(true) /\ all(Z) ) ) ) :- create_atom(Z).
macro( eg(F), nu( Z, F /\ ( all(false) \/ ex(Z) ) ) ) :- create_atom(Z).
macro( ef(F), mu( Z, F \/ ex(Z) ) ) :- create_atom(Z).

macro( h(A), mu( Z, ex(true) /\ allbut(A, Z) ) ) :- create_atom(Z).

macro( h(_,0), false ).
macro( h(A,N), ex(true) /\ allbut(A, h(A,N1) ) ) :- N > 0, N1 is N-1. 

macro( until(F,G), mu( Z, G \/ ( F /\ ex(true) /\ all(Z) ) ) ) :-
    create_atom( Z ).

macro( unless(F,G), nu( Z, G \/ ( F /\ all(Z) ) ) ) :-
    create_atom( Z ).

macro( ifthen(If,Then), ~If \/ Then ).
macro( ifthenelse(If,Then,Else), (If /\ Then) \/ (~If /\ Else ) ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% generic helper predicates
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic( create_atom_counter/1 ).

% create a new atom
create_atom( A ) :-
    ( retract( create_atom_counter(C) ),! ; C = 0 ),
    C1 is C + 1,
    asserta( create_atom_counter(C1) ),
    number_chars( C1, NumAtoms ),
    concat_atom_list( [z|NumAtoms], A ).

concat_atom_list( [X], X ) :- !.
concat_atom_list( [X|Xs], A ) :-
    concat_atom_list( Xs, RestA ),
    atom_concat( X, RestA, A ).

% map( X, [x1/y1, x2/y2, ...], Y ): map x to y in given mapping
map( X, [X/Y|_], Y ).
map( X, [_|M], Y ) :- map( X, M, Y ).

occurred( U, P ) :-
    occurred_u( U, P1 ), proc_equiv( P1, P ).

% subst for lists (see below)
subst( [], _, _, [] ) :- !.
subst( [F|Fs], V, Val, [F1|Fs1] ) :- !,
    subst( F, V, Val, F1 ),
    subst( Fs, V, Val, Fs1 ).

% subst( F, V, Val, F1 ) : F1 = F[V=Val], i. e. substitute Val for V in F
subst( V, V, Val, Val ) :- !.
subst( F, V, _, F ) :- atomic(F), F \= V, !.
subst( F, V, Val, F1 ) :- 
    compound(F),
    F =.. [Op|Arg],
    subst( Arg, V, Val, Arg1 ),
    F1 =.. [Op|Arg1].

% substall( F, S, F1 ) : F1 = F with all substitutions in S applied
substall( F, [], F ).
substall( F, [V/Def|T], F1 ) :-
    subst( F, V, Def, F2 ), substall( F2, T, F1 ).

% returns the number of states in the occurred list
numstates(N) :-
    findall( Z, occurred_u(Z,_), Zs ),
    list_length( Zs, N ).

list_length( [], 0 ).
list_length( [_|T], N ) :- list_length( T, N1 ), N is N1+1.

% uncomment this line if you don't want failure nodes printed
printfail( _,_,_ ) :- !.

% print failure node
printfail( F, P, S ) :-
    substall( F, S, F1 ),
    write( F1 ), write( ' does not hold in ' ), write( P ),nl.

